#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/ptrie.h"

typedef struct ptrie_node ptrie_node;

struct ptrie_node {
    unsigned int key_length;
    char *key;
    void *value;
    ptrie_value_free free;

    ptrie_node *next;
    ptrie_node *prev;
    ptrie_node *up;
    ptrie_node *down;
};

struct ptrie {
    ptrie_value_free free;
    ptrie_node *root;

    unsigned int length;
};

struct ptrie_iterator {
    ptrie *trie;
    ptrie_node *at;
};

ptrie_result ptrie_node_new(unsigned int, char *, void *, ptrie_value_free,
                            ptrie_node **);
ptrie_result ptrie_node_free(ptrie_node **);
ptrie_result ptrie_iterator_new(ptrie *,ptrie_iterator **);
static ptrie_result ptrie_set_above(ptrie_node *,unsigned int, char *, void *, ptrie_node **);
static ptrie_result ptrie_set_middle_prevnext(ptrie *,ptrie_node *,ptrie_node *,ptrie_node *,unsigned int , char *,void *);
static ptrie_result ptrie_node_shift_key(ptrie_node *,unsigned int);
static ptrie_result ptrie_node_prepend_key(ptrie_node *,unsigned int,char *);
/*static ptrie_result ptrie_print_kv(unsigned int,char *,unsigned int);*/
static ptrie_result ptrie_print_node(ptrie_node *,unsigned int);
static ptrie_result ptrie_print_node_at(ptrie_node *);
static ptrie_result ptrie_print_node_before(unsigned int,char *,char *,ptrie_node *);
static ptrie_result ptrie_print_node_after(ptrie_node *,char *why);

ptrie_result ptrie_node_new(unsigned int key_length, char *key,
                            void *value, ptrie_value_free free,ptrie_node **pthis) {
    ptrie_node *this;
    if(0 == key || 0 == pthis) {
        return PR_ERROR;
    }

    this = (ptrie_node *) malloc(
               sizeof(struct ptrie_node));
    this->key = (char *) malloc(sizeof(char) * key_length);

    strncpy(this->key, key, key_length);
    this->key_length = key_length;
    this->value = value;
    this->free=free;

    this->prev = 0;
    this->next = 0;
    this->up = 0;
    this->down = 0;

    *pthis = this;
    return PR_OK;
}

ptrie_result ptrie_node_free(ptrie_node **pthis) {
    ptrie_node *this;

    if (0 == pthis) {
        return PR_ERROR;
    }

    this = *pthis;
    if(0 == this) {
        return PR_ERROR;
    }
    if(0!=this->value && 0 != this->free) {
        (this->free)(this->value);
    }

    free(this->key);
    ptrie_node_free(&(this->next));
    ptrie_node_free(&(this->down));

    free(this);
    *pthis = 0;
    return PR_OK;
}

ptrie_result ptrie_iterator_new(ptrie *trie, ptrie_iterator **pthis) {
    ptrie_iterator *this = (ptrie_iterator *) malloc(
                               sizeof(struct ptrie_iterator));

    this->trie = trie;
    this->at = trie->root;
    *pthis = this;
    return PR_OK;
}

ptrie_result ptrie_iterator_free(ptrie_iterator **pthis) {
    ptrie_iterator *this;
    if (0 == pthis) {
        return PR_ERROR;
    }

    this = *pthis;
    if (0 == this) {
        return PR_ERROR;
    }

    free(this);
    *pthis = 0;
    return PR_OK;
}

ptrie_result ptrie_new(ptrie_value_free free, ptrie **pthis) {
    ptrie *this;
    if(0 == pthis || 0 == free) {
        return PR_ERROR;
    }

    this = (ptrie *) malloc(sizeof(struct ptrie));
    ptrie_node_new(0, "", 0, free, &(this->root));

    this->free = free;
    this->length=0;
    *pthis = this;
    return PR_OK;
}

ptrie_result ptrie_free(ptrie **pthis) {
    ptrie *this;
    if (0 == pthis) {
        return PR_ERROR;
    }

    this = *pthis;
    if(0 == this) {
        return PR_ERROR;
    }

    ptrie_node_free(&(this->root));
    free(this);
    *pthis = 0;
    return PR_OK;
}

ptrie_result ptrie_get_length(ptrie *this,unsigned int *plen) {
    if (0 == this) {
        return PR_ERROR;
    }

    *plen=this->length;
    return PR_OK;
}

static ptrie_result ptrie_set_above(ptrie_node *down,unsigned int key_length, char *key,
                                    void *value, ptrie_node **panew) {
    ptrie_node *anew;

    ptrie_node_new(key_length, key, value, down->free, &anew);

    anew->up=down->up;
    anew->down=down;
    anew->prev = down->prev;
    anew->next = down->next;

    if(0 != anew->next) {
        anew->next->prev=anew;
    }

    if(0 != anew->prev) {
        anew->prev->next=anew;
    } else {
        anew->up->down=anew;
    }

    down->up=anew;
    down->next=0;
    down->prev=0;

    if(0 != panew) {
        *panew = anew;
    }
    return PR_OK;
}


static ptrie_result ptrie_set_middle_prevnext(ptrie *trie,ptrie_node *up, ptrie_node *prev,ptrie_node *next,unsigned int key_length, char *key,
        void *value) {
    ptrie_node *anew;

    ptrie_node_new(key_length, key, value, trie->free, &anew);

    anew->prev=prev;
    anew->next=next;
    if(0 != next) {
        next->prev=anew;
        next->up=0;
    }
    if(0 != prev) {
        prev->next=anew;
    } else {
        anew->up=up;
        anew->up->down=anew;
    }
    return PR_OK;
}

static ptrie_result ptrie_node_shift_key(ptrie_node *this,unsigned int pos) {
    char *nkey;

    nkey=(char *)malloc(sizeof(char)*(this->key_length-pos));
    strncpy(nkey, this->key+pos, this->key_length-pos);
    free(this->key);

    this->key=nkey;
    this->key_length-=pos;
    return PR_OK;
}

static ptrie_result ptrie_node_prepend_key(ptrie_node *this,unsigned int len,char *pre) {
    char *nkey;

    nkey=(char *)malloc(sizeof(char)*(this->key_length+len));
    strncpy(nkey, pre, len);
    strncpy(nkey+len, this->key, this->key_length);
    free(this->key);

    this->key=nkey;
    this->key_length+=len;
    return PR_OK;
}

ptrie_result ptrie_set(ptrie *this, unsigned int key_length, char *key,
                       void *value) {
    ptrie_node *at;
    ptrie_node *up;
    ptrie_node *middle;
    unsigned int common_key_length;

    if (0 == this || 0 == key || 0 == key_length || 0 == value) {
        return PR_ERROR;
    }

    up = this->root;
    at = up->down;

    ptrie_print_node_before(key_length,key,"set",this->root);

    if(0 == at) {
        ptrie_set_middle_prevnext(this,up, 0, 0, key_length,key,value);
        ptrie_print_node_after(this->root, "added at root");
        this->length++;
        return PR_OK;
    }

    while (1) {
        ptrie_print_node_at(at);

        if (key[0] > at->key[0]) {
            if(0 == at->next) {
                ptrie_set_middle_prevnext(this,at->up, at, 0, key_length,key,value);
                ptrie_print_node_after(this->root, "added as last child");
                this->length++;
                return PR_OK;
            }
            at = at->next;
        } else if (key[0] < at->key[0]) {
            ptrie_set_middle_prevnext(this,up, at->prev, at, key_length, key, value);
            ptrie_print_node_after(this->root, "added before");
            this->length++;
            return PR_OK;
        } else {
            common_key_length = 1;
            while (common_key_length < key_length
                    && common_key_length < at->key_length) {

                if (key[ common_key_length]
                        != at->key[common_key_length]) {
                    break;
                }

                common_key_length++;
            }

            if (common_key_length == key_length && common_key_length == at->key_length) {
                (this->free)(at->value);
                at->value = value;

                ptrie_print_node_after(this->root, "just replaced value");
                return PR_OK;
            } else if(common_key_length == key_length) {
                ptrie_set_above(at, key_length, key, value, 0);
                ptrie_node_shift_key(at, common_key_length);

                ptrie_print_node_after(this->root, "in the middle");
                this->length++;
                return PR_OK;
            } else if(common_key_length == at->key_length) {
                key += common_key_length;
                key_length -= common_key_length;

                if(0 == at->down) {
                    ptrie_set_middle_prevnext(this,at, 0, 0, key_length,key,value);
                    ptrie_print_node_after(this->root, "first child");
                    this->length++;
                    return PR_OK;
                }
                up = at;
                at = up->down;
            } else {
                ptrie_set_above(at, common_key_length, key, 0, &middle);
                ptrie_node_shift_key(at, common_key_length);

                up = middle;
                key += common_key_length;
                key_length -= common_key_length;
            }
        }
    }
}

ptrie_result ptrie_get(ptrie *this, unsigned int key_length, char *key, void **pvalue) {
    ptrie_node *at;
    ptrie_node *up;
    unsigned int common_key_length;

    if (0 == this || 0 == key || 0 == key_length || 0 == pvalue) {
        return PR_ERROR;
    }

    up = this->root;
    at = up->down;

    ptrie_print_node_before(key_length,key,"get",this->root);

    while (0 != at && key_length > 0) {
        ptrie_print_node_at(at);
        if (key[0] > at->key[0]) {
            if(0 == at->next) {
                ptrie_print_node_after(this->root, "not found #1");
                return PR_ERROR;
            }
            at = at->next;
            continue;
        } else if (key[0] < at->key[0]) {
            ptrie_print_node_after(this->root, "not found #2");
            return PR_ERROR;
        }

        common_key_length = 1;
        while (common_key_length < key_length
                && common_key_length < at->key_length) {

            if (key[ common_key_length]
                    != at->key[common_key_length]) {
                break;
            }

            common_key_length++;
        }

        if(common_key_length == key_length && common_key_length == at->key_length) {
            if(0 == at->value) {
                ptrie_print_node_after(this->root, "found but empty");
                return PR_ERROR;
            }
            *pvalue = at->value;
            ptrie_print_node_after(this->root, "found");
            return PR_OK;
        }
        if(at->key_length != common_key_length) {
            ptrie_print_node_after(this->root, "not found #3");
            return PR_ERROR;
        }

        key_length -= at->key_length;
        key += at->key_length;
        up = at;
        at = up->down;
    }

    ptrie_print_node_after(this->root, "not found #4");
    return PR_ERROR;
}


ptrie_result ptrie_delete(ptrie *this, unsigned int key_length, char *key) {
    ptrie_node *at;
    ptrie_node *up;
    unsigned int common_key_length;

    if (0 == this || 0 == key || 0 == key_length) {
        return PR_ERROR;
    }

    up = this->root;
    at = up->down;
    ptrie_print_node_before(key_length,key,"delete",this->root);

    while (0 != at && key_length > 0) {
        ptrie_print_node_at(at);

        if (key[0] > at->key[0]) {
            if(0 == at->next) {
                ptrie_print_node_after(this->root, "not deleted, not found #1");
                return PR_ERROR;
            }
            at = at->next;
            continue;
        } else if (key[0] < at->key[0]) {
            ptrie_print_node_after(this->root, "not deleted, not found #2");
            return PR_ERROR;
        }

        common_key_length = 1;
        while (common_key_length < key_length
                && common_key_length < at->key_length) {

            if (key[ common_key_length]
                    != at->key[common_key_length]) {
                break;
            }

            common_key_length++;
        }

        if(common_key_length == key_length && common_key_length == at->key_length) {
            if(0 == at->value) {
                ptrie_print_node_after(this->root, "not deleted, no value, only node");
                return PR_ERROR;
            }

            (this->free)(at->value);
            at->value=0;

            if(0 != at->down) {
                if(0 != at->down->next) {
                    ptrie_print_node_after(this->root, "not removing, more than 1 children, but freed value");
                    this->length--;
                    return PR_OK;
                }

                ptrie_node_prepend_key(at->down, at->key_length, at->key);
                at->down->next=at->next;
                at->down->prev=at->prev;

                if(0 != at->next) {
                    at->next->prev=at->down;
                }

                if(0 != at->prev) {
                    at->prev->next=at->down;
                } else {
                    at->up->down=at->down;
                    at->down->up=at->up;
                }
            } else {
                if(0 != at->next) {
                    at->next->prev=at->prev;
                }

                if(0 != at->prev) {
                    at->prev->next=at->next;
                } else {
                    at->up->down=at->next;
                    if(0 != at->next) {
                        at->next->up=at->up;
                    }
                }
            }

            at->next=0;
            at->down=0;
            ptrie_node_free(&at);

            ptrie_print_node_after(this->root, "deleted");
            this->length--;
            return PR_OK;
        }

        if(common_key_length != at->key_length) {
            ptrie_print_node_after(this->root, "not deleted, not found #3");
            return PR_ERROR;
        }

        key_length -= at->key_length;
        key += at->key_length;
        up = at;
        at = up->down;
    }

    ptrie_print_node_after(this->root, "not deleted, not found #4");
    return PR_ERROR;
}

ptrie_result ptrie_iterate(ptrie *this, ptrie_iterator **pit) {
    if(0 == this || 0 == pit) {
        return PR_ERROR;
    }

    return ptrie_iterator_new(this, pit);
}

ptrie_result ptrie_iterator_next(ptrie_iterator *this) {
    if(0 == this || 0 == this->at) {
        return PR_ERROR;
    }

    while(0 != this->at) {
        if(0 != this->at->down) {
            this->at=this->at->down;
        } else if(0 != this->at->next) {
            this->at=this->at->next;
        } else {
            while(0 != this->at && 0 == this->at->next) {
                while(0!=this->at->prev) {
                    this->at=this->at->prev;
                }

                this->at=this->at->up;
            }

            if(0 != this->at) {
                this->at=this->at->next;
            }
        }

        if(0 != this->at && 0!=this->at->value) {
            return PR_OK;
        }
    }

    return PR_ERROR;

}

ptrie_result ptrie_iterator_get_key_length(ptrie_iterator *this,
        unsigned int *plen) {
    unsigned int length=0;
    ptrie_node *at;

    if(0==this || 0 == plen || 0 == this->at) {
        return PR_ERROR;
    }

    at=this->at;
    while(at!=this->trie->root) {
        length+=at->key_length;
        while(0!=at->prev) {
            at=at->prev;
        }
        at=at->up;
    }

    *plen=length;
    return PR_OK;
}

ptrie_result ptrie_iterator_copy_key(ptrie_iterator *this, char *key) {
    unsigned int from;
    unsigned int length=0;
    ptrie_node *at;

    if(0==this || 0 == key || 0 == this->at) {
        return PR_ERROR;
    }

    at=this->at;
    while(at!=this->trie->root) {
        length+=at->key_length;
        while(0!=at->prev) {
            at=at->prev;
        }
        at=at->up;
    }

    from=length;
    at=this->at;

    while(at!=this->trie->root) {
        from-=at->key_length;
        strncpy(key+from,at->key,at->key_length);

        while(0!=at->prev) {
            at=at->prev;
        }
        at=at->up;
    }

    return PR_OK;
}

ptrie_result ptrie_iterator_get_value(ptrie_iterator *this, void **pvalue) {
    if(0==this || 0 == pvalue || 0 == this->at) {
        return PR_ERROR;
    }

    *pvalue=this->at->value;
    return PR_OK;
}

/*
static ptrie_result ptrie_print_kv(unsigned int kl,char *k, unsigned int indent) {
    unsigned int i;

    for(i=0; i<indent; i++) {
        printf("    |");
    }
    printf(">");

    printf("%d <", kl);
    for(i=0; i<kl; i++) {
        printf("%c",k[i]);
    }
    printf(">\n");

    return PR_OK;
}
*/

static ptrie_result ptrie_print_node(ptrie_node *node,unsigned int indent) {
    return PR_OK;
    /*
    ptrie_node *it;

    if(0 == node) {
        return PR_OK;
    }

    ptrie_print_kv(node->key_length,node->key,indent);

    for(it=node->down; 0!=it; it=it->next) {
        ptrie_print_node(it,indent+1);
    }

    return PR_OK;
    */
}


static ptrie_result ptrie_print_node_before(unsigned int kl,char *k,char *operation,ptrie_node *this) {
    return PR_OK;
    /*
    printf("\nBEFORE %s on ", operation);
    ptrie_print_kv(kl,k,0);
    ptrie_print_node(this, 1);
    return PR_OK;
    */
}


static ptrie_result ptrie_print_node_after(ptrie_node *this, char *why) {
    return PR_OK;
    printf("AFTER, BECAUSE:     %s\n", why);
    ptrie_print_node(this, 1);
    printf("\n");
    return PR_OK;
}

static ptrie_result ptrie_print_node_at(ptrie_node *this) {
    return PR_OK;
    printf("... At:\n");
    return   ptrie_print_node(this, 2);
}


